from onetick.py.core.column_operations.base import _Operation
from onetick.py.types import value2str, nsectime, get_type_by_objects


def max(*objs):
    """
    Returns maximum value from list of ``objs``.
    The objects must be of the same type.

    Parameters
    ----------
    objs: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['MAX'] = otp.math.max(5, data['A'])
    >>> otp.run(data)
            Time  A  MAX
    0 2003-12-01  1    5
    """
    if len(objs) < 2:
        raise ValueError("otp.math.max expects at least 2 values to compare")

    def _max_func(*objs):
        dtype = get_type_by_objects(objs)
        onetick_params = map(value2str, objs)
        if dtype is nsectime:
            onetick_params = [f'NSECTIME_TO_LONG({param})' for param in onetick_params]
            onetick_params_str = ', '.join(onetick_params)
            return f"NSECTIME(MAX({onetick_params_str}))", dtype
        else:
            onetick_params_str = ', '.join(onetick_params)
            return f"MAX({onetick_params_str})", dtype

    return _Operation(
        op_func=_max_func,
        op_params=list(objs),
    )


def min(*objs):
    """
    Returns minimum value from list of ``objs``.
    The objects must be of the same type.

    Parameters
    ----------
    objs: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['MIN'] = otp.math.min(-5, data['A'])
    >>> otp.run(data)
            Time  A  MIN
    0 2003-12-01  1   -5
    """
    if len(objs) < 2:
        raise ValueError("otp.math.min expects at least 2 values to compare")

    def _min_func(*objs):
        dtype = get_type_by_objects(objs)
        onetick_params = map(value2str, objs)
        if dtype is nsectime:
            onetick_params = [f'NSECTIME_TO_LONG({param})' for param in onetick_params]
            onetick_params_str = ', '.join(onetick_params)
            return f"NSECTIME(MIN({onetick_params_str}))", dtype
        else:
            onetick_params_str = ', '.join(onetick_params)
            return f"MIN({onetick_params_str})", dtype

    return _Operation(
        op_func=_min_func,
        op_params=list(objs),
    )


def rand(min_value, max_value, seed=None):
    """
    Returns a pseudo-random value in the range between ``min_value`` and ``max_value`` (both inclusive).
    If ``seed`` is not specified, the function produces different values each time a query is invoked.
    If ``seed`` is specified, for this seed the function produces the same sequence of values
    each time a query is invoked.

    Parameters
    ----------
    min_value: int, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`
    max_value: int, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`
    seed: int, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['RAND'] = otp.math.rand(1, 1000)
    >>> otp.run(data)  # doctest: +SKIP
            Time  A  RAND
    0 2003-12-01  1   155
    """

    if isinstance(min_value, int) and min_value < 0:
        raise ValueError("It is not possible to use negative values for the `min_value`")
    if isinstance(min_value, int) and isinstance(max_value, int) and min_value >= max_value:
        raise ValueError("The `max_value` parameter should be more than `min_value`")

    def _random_func(min_value, max_value, seed=None):
        result = f'RAND({value2str(min_value)}, {value2str(max_value)}'
        if seed is not None:
            result += f', {value2str(seed)})'
        else:
            result += ')'
        return result, int

    return _Operation(
        op_func=_random_func,
        op_params=[min_value, max_value, seed],
    )


def frand(min_value=0, max_value=1, *, seed=None):
    """
    Returns a pseudo-random value in the range between ``min_value`` and ``max_value``.

    Parameters
    ----------
    min_value: float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`
    max_value: float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`
    seed: int, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`
        If not specified, the function produces different values each time a query is invoked.
        If specified, for this seed the function produces the same sequence of values
        each time a query is invoked.

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=otp.math.frand())
    >>> otp.run(data)  # doctest: +SKIP
            Time  A     FRAND
    0 2003-12-01  1  0.667519
    """

    if isinstance(min_value, (int, float)) and min_value < 0:
        raise ValueError("It is not possible to use negative values for the `min_value`")
    if isinstance(min_value, (int, float)) and isinstance(max_value, (int, float)) and min_value >= max_value:
        raise ValueError("The `max_value` parameter should be more than `min_value`")

    def _frand_func(min_value, max_value, seed=None):
        func_args = [min_value, max_value]
        if seed is not None:
            func_args.append(seed)
        onetick_args_str = ', '.join(value2str(arg) for arg in func_args)
        return f'FRAND({onetick_args_str})', float

    return _Operation(
        op_func=_frand_func,
        op_params=[min_value, max_value, seed],
    )


# TODO: this is not math, let's move it somewhere else
def now():
    """
    Returns the current time expressed as the number of milliseconds since the UNIX epoch in a GMT timezone.

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['NOW'] = otp.now()
    >>> otp.run(data)  # doctest: +SKIP
            Time  A                     NOW
    0 2003-12-01  1 2025-09-29 09:09:00.158
    """
    return _Operation(
        op_func=lambda: ('NOW()', nsectime),
    )


def ln(value):
    """
    Compute the natural logarithm of the ``value``.

    Parameters
    ----------
    value: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['LN'] = otp.math.ln(2.718282)
    >>> otp.run(data)
            Time  A   LN
    0 2003-12-01  1  1.0

    See Also
    --------
    :py:func:`onetick.py.math.exp`
    """
    return _Operation(
        op_func=lambda v: (f'LOG({value2str(v)})', float),
        op_params=[value],
    )


log = ln


def log10(value):
    """
    Compute the base-10 logarithm of the ``value``.

    Parameters
    ----------
    value: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['LOG10'] = otp.math.log10(100)
    >>> otp.run(data)
            Time  A  LOG10
    0 2003-12-01  1    2.0
    """
    return _Operation(
        op_func=lambda v: (f'LOG10({value2str(v)})', float),
        op_params=[value],
    )


def exp(value):
    """
    Compute the natural exponent of the ``value``.

    Parameters
    ----------
    value: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['E'] = otp.math.exp(1)
    >>> otp.run(data)
            Time  A         E
    0 2003-12-01  1  2.718282

    See Also
    --------
    :py:func:`onetick.py.math.ln`
    """
    return _Operation(
        op_func=lambda v: (f'EXP({value2str(v)})', float),
        op_params=[value],
    )


def sqrt(value):
    """
    Compute the square root of the ``value``.

    Parameters
    ----------
    value: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['SQRT'] = otp.math.sqrt(4)
    >>> otp.run(data)
            Time  A  SQRT
    0 2003-12-01  1   2.0
    """
    return _Operation(
        op_func=lambda v: (f'SQRT({value2str(v)})', float),
        op_params=[value],
    )


def sign(value):
    """
    Compute the sign of the ``value``.

    Parameters
    ----------
    value: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['SIGN_POS'] = otp.math.sign(123)
    >>> data['SIGN_ZERO'] = otp.math.sign(0)
    >>> data['SIGN_NEG'] = otp.math.sign(-123)
    >>> otp.run(data)
            Time  A  SIGN_POS  SIGN_ZERO  SIGN_NEG
    0 2003-12-01  1         1          0        -1
    """
    return _Operation(
        op_func=lambda v: (f'SIGN({value2str(v)})', int),
        op_params=[value],
    )


def pow(base, exponent):
    """
    Compute the ``base`` to the power of the ``exponent``.

    Parameters
    ----------
    base: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`
    exponent: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=2)
    >>> data['RES'] = otp.math.pow(data['A'], 10)
    >>> otp.run(data)
            Time  A     RES
    0 2003-12-01  2  1024.0
    """
    return _Operation(
        op_func=lambda b, e: (f'POWER({value2str(b)}, {value2str(e)})', float),
        op_params=[base, exponent],
    )


def pi():
    """
    Returns the value of Pi number.

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['PI'] = otp.math.pi()
    >>> otp.run(data)
            Time  A        PI
    0 2003-12-01  1  3.141593
    """
    return _Operation(
        op_func=lambda: ('PI()', float),
    )


def sin(value):
    """
    Returns the value of trigonometric function `sin` for the given ``value`` number expressed in radians.

    Parameters
    ----------
    value: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['SIN'] = otp.math.sin(otp.math.pi() / 6)
    >>> otp.run(data)
            Time  A  SIN
    0 2003-12-01  1  0.5

    See Also
    --------
    :py:func:`onetick.py.math.pi`
    :py:func:`onetick.py.math.asin`
    """
    return _Operation(
        op_func=lambda v: (f'SIN({value2str(v)})', float),
        op_params=[value],
    )


def cos(value):
    """
    Returns the value of trigonometric function `cos` for the given ``value`` number expressed in radians.

    Parameters
    ----------
    value: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['COS'] = otp.math.cos(otp.math.pi() / 3)
    >>> otp.run(data)
            Time  A  COS
    0 2003-12-01  1  0.5

    See Also
    --------
    :py:func:`onetick.py.math.pi`
    :py:func:`onetick.py.math.acos`
    """
    return _Operation(
        op_func=lambda v: (f'COS({value2str(v)})', float),
        op_params=[value],
    )


def tan(value):
    """
    Returns the value of trigonometric function `tan` for the given ``value`` number expressed in radians.

    Parameters
    ----------
    value: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['TAN'] = otp.math.tan(otp.math.pi() / 4)
    >>> otp.run(data)
            Time  A  TAN
    0 2003-12-01  1  1.0

    See Also
    --------
    :py:func:`onetick.py.math.pi`
    :py:func:`onetick.py.math.atan`
    """
    return _Operation(
        op_func=lambda v: (f'TAN({value2str(v)})', float),
        op_params=[value],
    )


def cot(value):
    """
    Returns the value of trigonometric function `cot` for the given ``value`` number expressed in radians.

    Parameters
    ----------
    value: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['COT'] = otp.math.cot(otp.math.pi() / 4)
    >>> otp.run(data)
            Time  A  COT
    0 2003-12-01  1  1.0

    See Also
    --------
    :py:func:`onetick.py.math.pi`
    :py:func:`onetick.py.math.acot`
    """
    return _Operation(
        op_func=lambda v: (f'COT({value2str(v)})', float),
        op_params=[value],
    )


def asin(value):
    """
    Returns the value of inverse trigonometric function `arcsin`.

    Parameters
    ----------
    value: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['ASIN'] = otp.math.asin(1).round(4)  # should return pi/2 ~ 1.5708
    >>> otp.run(data)
            Time  A    ASIN
    0 2003-12-01  1  1.5708

    `otp.math.arcsin()` is the alias for `otp.math.asin()`:

    >>> data = otp.Tick(A=1)
    >>> data['ASIN'] = otp.math.arcsin(1).round(4)
    >>> otp.run(data)
            Time  A    ASIN
    0 2003-12-01  1  1.5708

    See Also
    --------
    :py:func:`onetick.py.math.pi`
    :py:func:`onetick.py.math.sin`
    """
    return _Operation(
        op_func=lambda v: (f'ASIN({value2str(v)})', float),
        op_params=[value],
    )


arcsin = asin


def acos(value):
    """
    Returns the value of inverse trigonometric function `arccos`.

    Parameters
    ----------
    value: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['ACOS'] = otp.math.acos(-1).round(4)  # should return pi ~ 3.1416
    >>> otp.run(data)
            Time  A    ACOS
    0 2003-12-01  1  3.1416

    `otp.math.arccos()` is the alias for `otp.math.acos()`:

    >>> data = otp.Tick(A=1)
    >>> data['ACOS'] = otp.math.arccos(-1).round(4)
    >>> otp.run(data)
            Time  A    ACOS
    0 2003-12-01  1  3.1416

    See Also
    --------
    :py:func:`onetick.py.math.pi`
    :py:func:`onetick.py.math.cos`
    """
    return _Operation(
        op_func=lambda v: (f'ACOS({value2str(v)})', float),
        op_params=[value],
    )


arccos = acos


def atan(value):
    """
    Returns the value of inverse trigonometric function `arctan`.

    Parameters
    ----------
    value: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['ATAN'] = otp.math.atan(1).round(4)  # should return pi/4 ~ 0.7854
    >>> otp.run(data)
            Time  A    ATAN
    0 2003-12-01  1  0.7854

    `otp.math.arctan()` is the alias for `otp.math.atan()`:

    >>> data = otp.Tick(A=1)
    >>> data['ATAN'] = otp.math.arctan(1).round(4)
    >>> otp.run(data)
            Time  A    ATAN
    0 2003-12-01  1  0.7854

    See Also
    --------
    :py:func:`onetick.py.math.pi`
    :py:func:`onetick.py.math.tan`
    """
    return _Operation(
        op_func=lambda v: (f'ATAN({value2str(v)})', float),
        op_params=[value],
    )


arctan = atan


def acot(value):
    """
    Returns the value of inverse trigonometric function `arccot`.

    Parameters
    ----------
    value: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data['ACOT'] = otp.math.acot(1).round(4)  # should return pi/4 ~ 0.7854
    >>> otp.run(data)
            Time  A    ACOT
    0 2003-12-01  1  0.7854

    `otp.math.arccot()` is the alias for `otp.math.acot()`:

    >>> data = otp.Tick(A=1)
    >>> data['ACOT'] = otp.math.arccot(1).round(4)
    >>> otp.run(data)
            Time  A    ACOT
    0 2003-12-01  1  0.7854

    See Also
    --------
    :py:func:`onetick.py.math.pi`
    :py:func:`onetick.py.math.cot`
    """
    return _Operation(
        op_func=lambda v: (f'ACOT({value2str(v)})', float),
        op_params=[value],
    )


arccot = acot


def mod(value1, value2):
    """
    Computes the remainder from dividing ``value1`` by ``value2``.

    Parameters
    ----------
    value1: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`
    value2: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=100)
    >>> data['MOD'] = otp.math.mod(data['A'], 72)
    >>> otp.run(data)
            Time    A  MOD
    0 2003-12-01  100   28
    """
    return _Operation(
        op_func=lambda v1, v2: (f'MOD({value2str(v1)}, {value2str(v2)})', int),
        op_params=[value1, value2],
    )


def div(value1, value2):
    """
    Computes the quotient by dividing ``value1`` by ``value2``.

    Parameters
    ----------
    value1: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`
    value2: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=100)
    >>> data['DIV'] = otp.math.div(data['A'], 72)
    >>> otp.run(data)
            Time    A  DIV
    0 2003-12-01  100    1
    """
    return _Operation(
        op_func=lambda v1, v2: (f'DIV({value2str(v1)}, {value2str(v2)})', int),
        op_params=[value1, value2],
    )


def gcd(value1, value2):
    """
    Computes the greatest common divisor between ``value1`` and ``value2``.

    Parameters
    ----------
    value1: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`
    value2: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Tick(A=99)
    >>> data['GCD'] = otp.math.gcd(data['A'], 72)
    >>> otp.run(data)
            Time   A  GCD
    0 2003-12-01  99    9
    """
    return _Operation(
        op_func=lambda v1, v2: (f'GCD({value2str(v1)}, {value2str(v2)})', int),
        op_params=[value1, value2],
    )


def floor(value):
    """
    Returns a float value representing the largest number that is less than or equal to the ``value``.

    Note
    ----
    Rounding :class:`otp.nan <onetick.py.nan>` returns NaN
    and rounding :class:`otp.inf <onetick.py.inf>` returns Infinity.

    Parameters
    ----------
    value: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Ticks(A=[-1.7, -1.5, -1.2, -1, 0 , 1, 1.2, 1.5, 1.7, -otp.inf, otp.inf, otp.nan])
    >>> data['FLOOR'] = otp.math.floor(data['A'])
    >>> otp.run(data)
                          Time     A  FLOOR
    0  2003-12-01 00:00:00.000  -1.7   -2.0
    1  2003-12-01 00:00:00.001  -1.5   -2.0
    2  2003-12-01 00:00:00.002  -1.2   -2.0
    3  2003-12-01 00:00:00.003  -1.0   -1.0
    4  2003-12-01 00:00:00.004   0.0    0.0
    5  2003-12-01 00:00:00.005   1.0    1.0
    6  2003-12-01 00:00:00.006   1.2    1.0
    7  2003-12-01 00:00:00.007   1.5    1.0
    8  2003-12-01 00:00:00.008   1.7    1.0
    9  2003-12-01 00:00:00.009  -inf   -inf
    10 2003-12-01 00:00:00.010   inf    inf
    11 2003-12-01 00:00:00.011   NaN    NaN
    """
    def fun(v):
        v = value2str(v)
        return f'CASE({v}, NAN(), NAN(), INFINITY(), INFINITY(), -INFINITY(), -INFINITY(), FLOOR({v}) * 1.0)', float

    return _Operation(
        op_func=fun,
        op_params=[value],
    )


def ceil(value):
    """
    Returns a float value representing the smallest number that is greater than or equal to the ``value``.

    Note
    ----
    Rounding :class:`otp.nan <onetick.py.nan>` returns NaN
    and rounding :class:`otp.inf <onetick.py.inf>` returns Infinity.

    Parameters
    ----------
    value: int, float, :py:class:`~onetick.py.Operation`, :py:class:`~onetick.py.Column`

    Returns
    -------
        :py:class:`~onetick.py.Operation`

    Examples
    --------
    >>> data = otp.Ticks(A=[-1.7, -1.5, -1.2, -1, 0 , 1, 1.2, 1.5, 1.7, -otp.inf, otp.inf, otp.nan])
    >>> data['CEIL'] = otp.math.ceil(data['A'])
    >>> otp.run(data)
                          Time     A  CEIL
    0  2003-12-01 00:00:00.000  -1.7  -1.0
    1  2003-12-01 00:00:00.001  -1.5  -1.0
    2  2003-12-01 00:00:00.002  -1.2  -1.0
    3  2003-12-01 00:00:00.003  -1.0  -1.0
    4  2003-12-01 00:00:00.004   0.0   0.0
    5  2003-12-01 00:00:00.005   1.0   1.0
    6  2003-12-01 00:00:00.006   1.2   2.0
    7  2003-12-01 00:00:00.007   1.5   2.0
    8  2003-12-01 00:00:00.008   1.7   2.0
    9  2003-12-01 00:00:00.009  -inf  -inf
    10 2003-12-01 00:00:00.010   inf   inf
    11 2003-12-01 00:00:00.011   NaN   NaN
    """
    def fun(v):
        v = value2str(v)
        return f'CASE({v}, NAN(), NAN(), INFINITY(), INFINITY(), -INFINITY(), -INFINITY(), CEIL({v}) * 1.0)', float

    return _Operation(
        op_func=fun,
        op_params=[value],
    )


def round(value, precision=0, rounding_method='upward'):
    """
    Rounds value with specified ``precision`` and ``rounding_method``.

    Rounding :class:`otp.nan <onetick.py.nan>` returns NaN
    and rounding :class:`otp.inf <onetick.py.inf>` returns Infinity.

    Parameters
    ----------
    precision: int
        Number from -12 to 12.
        Positive precision is precision after the floating point.
        Negative precision is precision before the floating point (and the fraction part is dropped in this case).
    rounding_method: str
        Used for values that are exactly half-way between two integers (when the fraction part of value is exactly 0.5).
        Available values are **upward**, **downward**, **towards_zero**, **away_from_zero**.
        Default is **upward**.

    Examples
    --------

    Different rounding methods produce different results for values that are exactly half-way between two integers:

    >>> t = otp.Ticks(A=[-123.45, 123.45, -123.4, 123.6])
    >>> t['UPWARD'] = otp.math.round(t['A'], precision=1, rounding_method='upward')
    >>> t['DOWNWARD'] = otp.math.round(t['A'], precision=1, rounding_method='downward')
    >>> t['TOWARDS_ZERO'] = otp.math.round(t['A'], precision=1, rounding_method='towards_zero')
    >>> t['AWAY_FROM_ZERO'] = otp.math.round(t['A'], precision=1, rounding_method='away_from_zero')
    >>> otp.run(t).head(2)
                         Time       A  UPWARD  DOWNWARD  TOWARDS_ZERO  AWAY_FROM_ZERO
    0 2003-12-01 00:00:00.000 -123.45  -123.4    -123.5        -123.4          -123.5
    1 2003-12-01 00:00:00.001  123.45   123.5     123.4         123.4           123.5

    Note that for other cases all methods produce the same results:

    >>> otp.run(t).tail(2)
                         Time      A  UPWARD  DOWNWARD  TOWARDS_ZERO  AWAY_FROM_ZERO
    2 2003-12-01 00:00:00.002 -123.4  -123.4    -123.4        -123.4          -123.4
    3 2003-12-01 00:00:00.003  123.6   123.6     123.6         123.6           123.6

    Positive precision truncates to the number of digits *after* floating point:

    >>> t = otp.Ticks(A=[-123.45, 123.45])
    >>> t['ROUND1'] = otp.math.round(t['A'], 1)
    >>> t['ROUND2'] = otp.math.round(t['A'], 2)
    >>> otp.run(t)
                            Time       A  ROUND1  ROUND2
    0 2003-12-01 00:00:00.000 -123.45  -123.4 -123.45
    1 2003-12-01 00:00:00.001  123.45   123.5  123.45

    Negative precision truncates to the number of digits *before* floating point
    (and the fraction part is dropped in this case):

    >>> t = otp.Ticks(A=[-123.45, 123.45])
    >>> t['ROUND_M1'] = otp.math.round(t['A'], -1)
    >>> t['ROUND_M2'] = otp.math.round(t['A'], -2)
    >>> otp.run(t)
                            Time       A  ROUND_M1  ROUND_M2
    0 2003-12-01 00:00:00.000 -123.45    -120.0    -100.0
    1 2003-12-01 00:00:00.001  123.45     120.0     100.0

    Rounding :class:`otp.nan <onetick.py.nan>` returns NaN
    and rounding :class:`otp.inf <onetick.py.inf>` returns Infinity in all cases:

    >>> t = otp.Ticks(A=[otp.inf, -otp.inf, otp.nan])
    >>> t['ROUND_0'] = otp.math.round(t['A'])
    >>> t['ROUND_P2'] = otp.math.round(t['A'], 2)
    >>> t['ROUND_M2'] = otp.math.round(t['A'], -2)
    >>> otp.run(t)
                         Time    A  ROUND_0  ROUND_P2  ROUND_M2
    0 2003-12-01 00:00:00.000  inf      inf       inf       inf
    1 2003-12-01 00:00:00.001 -inf     -inf      -inf      -inf
    2 2003-12-01 00:00:00.002  NaN      NaN       NaN       NaN
    """
    if not -12 <= precision <= 12:
        raise ValueError("Parameter 'precision' must be an integer in range [-12, 12]")
    supported_rounding_methods = {'upward', 'downward', 'towards_zero', 'away_from_zero'}
    if rounding_method not in supported_rounding_methods:
        raise ValueError(f"Parameter 'rounding_method' must be one of {supported_rounding_methods}")
    return _Operation(
        op_func=lambda v, p, r: (f'ROUND_DOUBLE({value2str(v)},{value2str(p)},{value2str(r)})', float),
        op_params=[value, precision, rounding_method],
    )



# pylama:ignore=W0237
import logging
import os
import datetime as dt
import subprocess
import warnings

from datetime import timedelta
from collections import defaultdict
from dateutil.relativedelta import relativedelta
from typing import List, Union, Optional
from uuid import uuid4

from onetick import py as otp
from onetick.py.core import db_constants as constants
from onetick.py.compatibility import is_native_plus_zstd_supported
from onetick.py import utils, sources, session, configuration

import pandas


def _tick_type_detector(tick_type, obj):
    if tick_type is not None:
        return tick_type

    type2traits = {
        "ORDER": ["ID", "BUY_FLAG", "SIDE", "QTY", "QTY_FILLED", "QTY", "ORDTYPE", "PRICE", "PRICE_FILLED"],
        "QTE": ["ASK_PRICE", "BID_PRICE", "ASK_SIZE", "BID_SIZE"],
        "NBBO": ["ASK_PRICE", "BID_PRICE", "ASK_SIZE", "BID_SIZE", "BID_EXCHANGE", "ASK_EXCHANGE"],
        "TRD": ["PRICE", "SIZE"],
    }

    type2count = defaultdict(lambda: 0)
    max_tt = "TRD"
    max_count = 0

    for column in obj.columns():
        for tt, traits in type2traits.items():
            if column in traits:
                type2count[tt] += 1

                if type2count[tt] > max_count:
                    max_count = type2count[tt]
                    max_tt = tt

    return max_tt


def write_to_db(src: 'otp.Source',
                dest: Union[str, 'otp.DB'],
                date: dt.date,
                symbol: Union[str, 'otp.Column'],
                tick_type: Union[str, 'otp.Column'],
                timezone: Optional[str] = None,
                execute: bool = True,
                start: Optional[dt.date] = None,
                end: Optional[dt.date] = None,
                propagate: bool = False,
                append: bool = True,
                **kwargs):
    """
    Writes source to the database.
    The main differences from otp.Source.write() function are
    appending ticks by default, automatic tick_type detection and executing the query right here.

    Parameters
    ----------
    src: :class:`otp.Source`
        source that will be written to the database.
    dest: str or :py:class:`otp.DB <onetick.py.DB>`
        database name or object.
    date: datetime or None
        date where to save data.
        Cannot be used together with `start` and `end` parameters.
    start: datetime or None
        start date of period where to save data.
        Cannot be used together with `date` parameters.
        Default is None.
    end: datetime or None
        end date of period where to save data.
        Cannot be used together with `date` parameters.
        Default is None.
    symbol: str or Column
        resulting symbol name string or column to get symbol name from.
    tick_type: str or Column
        resulting tick type string or column to get tick type from.
        If tick type is None then an attempt will be taken to get
        tick type name automatically based on the ``src`` source's schema.
        (ORDER, QTE, TRD and NBBO tick types are supported).
    timezone: str
        If ``execute`` parameter is set then this timezone
        will be used for running the query.
        By default, it is set to `otp.config.tz`.
    execute: bool
        execute the query right here or not.
        If True, `date` or `start`+`end` parameters are required,
        and then dataframe will be returned.
        (Probably empty, unless 'propagate' parameter is specified).
        If False, modified copy of the source ``src`` will be returned.
    propagate: bool
        Propagate ticks after writing or not.
    append: bool
        Write the data in append mode or not.
    kwargs:
        other arguments that will be passed to :py:meth:`onetick.py.Source.write` function.
    """

    tick_type = _tick_type_detector(tick_type, src)

    if timezone is None:
        timezone = configuration.config.tz
    if date is None or date is otp.adaptive:
        kwargs['start_date'] = start
        kwargs['end_date'] = end
    else:
        kwargs['date'] = date

    writer = src.write(db=dest,
                       symbol=symbol,
                       tick_type=tick_type,
                       propagate=propagate,
                       append=append,
                       **kwargs)

    if execute:
        if not start or not end:
            start = getattr(date, 'ts', date)
            end = start + relativedelta(days=1)
        return otp.run(writer, start=start, end=end, timezone=timezone)
    return writer


class _DB:

    _LOCAL = False
    # this flag means that db section should be added to locator and acl (True)
    # or db locates on some ts (False)

    def __init__(
        self,
        name,
        src=None,
        date=None,
        symbol=None,
        tick_type=None,
        db_properties: Optional[dict] = None,
        db_locations: Optional[list] = None,
        db_raw_data: Optional[dict] = None,
        db_feed: Optional[dict] = None,
        write=True,
        minimum_start_date: Optional[Union[str, dt.date, 'otp.date']] = None,
        maximum_end_date: Optional[Union[str, dt.date, 'otp.date']] = None,
    ):
        # we assume here that db_properties and db_locations fully prepared here or None (in case of remote db)
        self.name = name
        self.id, _, _ = name.partition("//")
        self._db_properties = db_properties
        self._db_locations = db_locations
        self._db_raw_data = db_raw_data
        self._db_feed = db_feed
        self._write = write
        self._minimum_start_date = (
            dt.datetime.strptime(minimum_start_date, '%Y%m%d').date()
            if isinstance(minimum_start_date, str)
            else minimum_start_date
        )
        self._maximum_end_date = (
            dt.datetime.strptime(maximum_end_date, '%Y%m%d').date()
            if isinstance(maximum_end_date, str)
            else maximum_end_date
        )
        if src is not None:
            self.add(src=src, date=date, symbol=symbol, tick_type=tick_type)
        elif any(x is not None for x in (date, symbol, tick_type)):
            warnings.warn(
                "Parameters 'date', 'symbol' and 'tick_type' can only be set when parameter 'src' is specified.",
                FutureWarning,
                stacklevel=3,
            )

    @staticmethod
    def _format_params(params):
        res = {}
        for key, value in params.items():
            if hasattr(value, 'strftime'):
                res[key] = value.strftime("%Y%m%d%H%M%S")
            else:
                res[key] = str(value)
        return res

    # TODO: move this method to DB
    def add(self,
            src,
            date=None,
            start=None,
            end=None,
            symbol=None,
            tick_type=None,
            timezone=None,
            **kwargs):
        """
        Add data to a database.
        If ticks with the same timestamp are already presented in database old values won't be updated.

        Parameters
        ----------
        src: :class:`otp.Source`
            source that will be written to the database.
        date: datetime or None
            date of the day in which the data will be saved.
            The timestamps of the ticks should be between the start and the end of the day.
            Be default, it is set to `otp.config.default_date`.
        start: datetime or None
            start day of period in which the data will be saved.
            The timestamps of the ticks should be between `start` and `end` dates.
            Cannot be used with `date` parameter.
            Be default, None.
        end: datetime or None
            end day of period in which the data will be saved.
            The timestamps of the ticks should be between `start` and `end` dates.
            Cannot be used with `date` parameter.
            Be default, None.
        symbol: str or Column
            resulting symbol name string or column to get symbol name from.
            Be default, it is set to `otp.config.default_db_symbol`.
        tick_type: str or Column
            resulting tick type string or column to get tick type from.
            If tick type is None then an attempt will be taken to get
            tick type name automatically based on the ``src`` source's schema.
            (ORDER, QTE, TRD and NBBO tick types are supported).
        timezone: str
            This timezone will be used for running the query.
            By default, it is set to `otp.config.tz`.
        kwargs:
            other arguments that will be passed to :py:meth:`onetick.py.Source.write` function.

        Examples
        --------

        Data is saved to the specified date, symbol and tick type:
        (note that ``session`` is created before this example)

        >>> db = otp.DB('MYDB2')
        >>> db.add(otp.Ticks(A=[4, 5, 6]), date=otp.dt(2003, 1, 1), symbol='SMB', tick_type='TT')
        >>> session.use(db)

        We can get the same data by specifying the same parameters:

        >>> data = otp.DataSource(db, date=otp.dt(2003, 1, 1), symbols='SMB', tick_type='TT')
        >>> otp.run(data)
                             Time  A
        0 2003-01-01 00:00:00.000  4
        1 2003-01-01 00:00:00.001  5
        2 2003-01-01 00:00:00.002  6
        """
        if timezone is None:
            timezone = configuration.config.tz
        if start and end and date:
            raise ValueError("You can't specify both start/end and date parameters")

        if start and end:
            kwargs['start'] = start
            kwargs['end'] = end
            kwargs['date'] = otp.adaptive
        else:
            kwargs['date'] = date if date is not None else configuration.config.default_start_time

        _symbol = symbol if symbol is not None else configuration.config.default_db_symbol
        kwargs.setdefault('propagate', kwargs.get('propagate_ticks', False))

        res = self._session_handler(write_to_db,
                                    src=src,
                                    dest=self.name,
                                    symbol=_symbol,
                                    tick_type=tick_type,
                                    timezone=timezone,
                                    **kwargs)

        # We need to keep backward-compatibility,
        # because before there was no ability to get written ticks
        if kwargs.get('propagate'):
            return res
        else:
            return None

    @property
    def properties(self):
        """
        Get dict of database properties.

        Returns
        -------
        dict

        Examples
        --------
        >>> otp.DB('X').properties  # doctest: +SKIP
        {'symbology': 'BZX',
         'archive_compression_type': 'NATIVE_PLUS_GZIP',
         'tick_timestamp_type': 'NANOS'}
        """
        return self._db_properties

    @property
    def locations(self):
        """
        Get list of database locations.

        Returns
        -------
        list of dict

        Examples
        --------
        >>> otp.DB('X').locations # doctest:+ELLIPSIS
        [{'access_method': 'file',
          'start_time': '20021230000000',
          'end_time': '21000101000000',
          ...}]
        """
        return self._db_locations

    @property
    def raw_data(self):
        """
        Get dict of database raw configurations.

        Returns
        -------
        dict of dict

        Examples
        --------
        >>> db = otp.DB('RAW_EXAMPLE',
        ...     db_raw_data=[{
        ...         'id': 'PRIMARY_A',
        ...         'prefix': 'DATA.',
        ...         'locations': [
        ...             {'mount': 'mount1'}
        ...         ]
        ...     }]
        ... )
        >>> db.raw_data # doctest:+ELLIPSIS
        [{'id': 'PRIMARY_A', 'prefix': 'DATA.', 'locations': [{'mount': 'mount1', 'access_method': 'file', ...}]}]
        """
        return self._db_raw_data

    @property
    def feed(self):
        """
        Get dict of database feed configuration.

        Returns
        -------
        dict

        Examples
        --------
        >>> db = otp.DB('RAW_EXAMPLE',
        ...     db_raw_data=[{
        ...         'id': 'PRIMARY_A',
        ...         'prefix': 'DATA.',
        ...         'locations': [
        ...             {'mount': 'mount1'}
        ...         ]
        ...     }],
        ...     db_feed={'type': 'rawdb', 'raw_source': 'PRIMARY_A'},
        ... )
        >>> db.feed
        {'type': 'rawdb', 'raw_source': 'PRIMARY_A', 'format': 'native'}
        """
        return self._db_feed

    @property
    def symbols(self):
        result = self._session_handler(self._symbols)
        return result if result else []

    def _session_handler(self, func, *args, **kwargs):
        """
        Handler to check if database is already in locator and
        run function with separate session or using current

        :param func: function to run
        """
        __result = None

        _session = session.Session._instance
        close_session = False
        _remove_from_locator = False
        _remove_from_acl = False
        if _session is None:
            close_session = True
            _session = session.Session()

        try:
            if self._LOCAL:
                if self.id not in _session.locator.databases:
                    if self.id != self.name and not otp.compatibility.is_supported_reload_locator_with_derived_db():
                        # Derived DB
                        raise ValueError(
                            "You need include derived DB into the session use the .use method before adding "
                            " data there."
                        )

                    _remove_from_locator = True
                    _session.locator.add(self)
                if self.id not in _session.acl.databases:
                    _remove_from_acl = True
                    _session.acl.add(self)
            __result = func(*args, **kwargs)
        finally:
            if close_session:
                _session.close()
            else:
                if self._LOCAL:
                    if _remove_from_locator:
                        _session.locator.remove(self)
                    if _remove_from_acl:
                        _session.acl.remove(self)

        return __result

    def _symbols(self):
        src = sources.Symbols(self)
        symbols = otp.run(src)
        result = []
        if symbols.empty:
            return []
        for s in list(symbols["SYMBOL_NAME"]):
            result.append(s.split(":")[-1])
        return result

    def __repr__(self):
        return "DB: " + self.name

    def __str__(self):
        return self.name


class DB(_DB):
    """
    A class representing OneTick databases when configuring
    :py:class:`locators <onetick.py.session.Locator>` and :py:class:`ACL <onetick.py.session.ACL>`.

    A database can point to an existing OneTick archive
    or the temporary directory can be created with the data provided.

    This class object can then be :py:meth:`used <onetick.py.Session.use>`
    in :py:class:`onetick.py.Session`.

    The data to add to the database can be passed into the
    constructor with parameter ``src`` or later with the :py:meth:`add` method.

    Note that after ticks were written to a particular timestamps in the database,
    they can't be updated for the same timestamps.

    Note
    ----
    This class can only be used to create database on the local machine.
    Database is created by using/creating a directory in the local filesystem
    and adding entries to the OneTick locator and ACL local files.
    This class can't be used to manage remote databases.

    Parameters
    ----------
    name : str
        Database name.
        Derived databases are specified in "parent//derived" format.
        A derived database inherits the parent database properties.
    src : :py:class:`~onetick.py.Source`, optional
        Data to add to the database.
    date : :py:class:`onetick.py.date` or :py:class:`datetime.date`, optional
        ``src`` data will be added to this date.
        Can be set only if ``src`` is set.
        Default value is the same as in :py:meth:`add` method.
    symbol : str, optional
        Symbol name to add ``src`` data for.
        Can be set only if ``src`` is set.
        Default value is the same as in :py:meth:`add` method.
    tick_type : str, optional
        Tick type to add ``src`` data for.
        Can be set only if ``src`` is set.
        Default value is the same as in :py:meth:`add` method.
    db_properties : :obj:`dict`, optional
        Properties of the database to add to the locator
    db_locations : :obj:`list` of :obj:`dict`, optional
        A list of locations for the database to add to the locator.
        This parameter is a list, because databases in a locator can have several location sections.
        If not specified, a temporary directory is used as the database location.
    db_raw_data: :obj:`list` of :obj:`dict`, optional
        Raw database configuration.
    db_feed: dict, optional
        Feed configuration.
    write : bool, optional
        Flag that controls access to write to database.
    clean_up : bool, optional
        Flag that controls temporary database cleanup
    destroy_access : bool, optional
        Flag that controls access to destroy the database.
    minimum_start_date: str or :py:class:`datetime.date` or :py:class:`onetick.py.date`
        Specifies the minimum date of the tick that can be served to a user.
        The format for the value is *YYYYMMDD*.
        The OneTick server enforces this permission by verifying that
        the query start time is not less than time of 00:00:00, in GMT timezone, for the specified minimum start date.
    maximum_end_date: str or :py:class:`datetime.date` or :py:class:`onetick.py.date`
        Specifies the date of the tick, starting from which ticks are not allowed to be returned to a user.
        The format for the value is *YYYYMMDD*.
        The OneTick server enforces this permission by verifying that
        the query end time is less than time of 00:00:00, in GMT timezone, for the specified maximum end date.

    Examples
    --------

    A database can be initialized along with data:

    >>> data = otp.Ticks(X=['hello', 'world!'])
    >>> db = otp.DB('MYDB', data)

    Database symbol name, tick type and date to which the data will be written can be changed like this:

    >>> db = otp.DB('MYDB', data, symbol='S_S', tick_type='T_T', date=otp.dt(2003, 12, 1))

    You can specify a derived db by using ``//`` as a separator:

    >>> data = otp.Ticks(X=['parent1', 'parent2'])
    >>> db = otp.DB('DB_A', data)
    >>> db.add(data)

    >>> data_derived = otp.Ticks(X=['derived1', 'derived2'])
    >>> db_derived = otp.DB('DB_A//DB_D')
    >>> session.use(db_derived)
    >>> db_derived.add(data_derived)

    You can add an existing OneTick database to the locator or create a new one:

    >>> existing_db = otp.DB('MY_US_COMP',  # doctest: +SKIP
    ...                      db_locations=[{'location': '/home/user/data/US_COMP',
    ...                                     'start_time': datetime(2003, 1, 1),
    ...                                     'end_time': datetime(2010, 1, 1),
    ...                                     'day_boundary_tz': 'EST5EDT'}])
    >>> session.use(existing_db)  # doctest: +SKIP
    """

    _LOCAL = True

    def __init__(
        self,
        name=None,
        src=None,
        date=None,
        symbol=None,
        tick_type=None,
        kind='archive',
        db_properties=None,
        db_locations=None,
        db_raw_data=None,
        db_feed=None,
        write=True,
        clean_up=utils.default,
        destroy_access=False,
        minimum_start_date=None,
        maximum_end_date=None,
    ):
        if name is not None and not isinstance(name, str):
            message = f"Database name expected to be string got {type(name)}"
            logging.error(message)
            raise TypeError(message)

        self._clean_up = clean_up
        self._destroy_access = destroy_access
        self._path = None
        self._db_suffix = ""

        if name:
            self._db_suffix = name
        else:
            # Mostly for temporary databases
            name = uuid4().hex.upper()
            self._db_suffix = "db_" + name

        db_properties = self._prepare_db_properties(db_properties)
        db_day_boundary_tz_set = 'day_boundary_tz' in db_properties.keys()
        db_locations = self._prepare_db_locations(db_locations,
                                                  db_day_boundary_tz_set=db_day_boundary_tz_set,
                                                  kind=kind)
        db_raw_data = self._prepare_db_raw_data(db_raw_data, db_properties)
        db_feed = self._prepare_db_feed(db_feed)

        if isinstance(src, pandas.DataFrame):
            csv_path = os.path.join(self._tmp_dir.path, uuid4().hex.upper() + ".csv")
            src.to_csv(csv_path, index=False)
            if otp.__webapi__:
                src = sources.CSV(utils.FileBuffer(csv_path))
            else:
                src = sources.CSV(csv_path)

        super().__init__(
            name=name,
            src=src,
            date=date,
            symbol=symbol,
            tick_type=tick_type,
            db_properties=db_properties,
            db_locations=db_locations,
            db_raw_data=db_raw_data,
            db_feed=db_feed,
            write=write,
            minimum_start_date=minimum_start_date,
            maximum_end_date=maximum_end_date,
        )

    def _prepare_db_properties(self, properties):
        if properties is None:
            properties = {}

        # convert all property keys to lowercase
        properties = {k.lower(): v for k, v in properties.items()}

        # set default properties if they are not specified
        properties.setdefault("symbology", configuration.config.default_symbology)
        properties.setdefault("tick_timestamp_type", "NANOS")

        if is_native_plus_zstd_supported():
            properties.setdefault("archive_compression_type", constants.compression_type.NATIVE_PLUS_ZSTD)
        else:
            properties.setdefault("archive_compression_type", constants.compression_type.NATIVE_PLUS_GZIP)

        return self._format_params(properties)

    def _create_db(self):
        logging.debug(f'Creating temporary directory for db "{self._db_suffix}"')

        dirs_list = self._db_suffix.replace("//", " DERIVED ").split()
        dir_name = ''
        base_dir = ""
        if os.getenv('OTP_WEBAPI_TEST_MODE'):
            # copied from onetick.test.fixtures _keep_generated_dir()
            base_dir = os.path.join(otp.utils.TMP_CONFIGS_DIR(), os.environ.get("ONE_TICK_TMP_DIR", "dbs"))
        for cur_dir in dirs_list:
            dir_name = os.path.join(dir_name, cur_dir)
            self._tmp_dir = utils.TmpDir(dir_name, clean_up=self._clean_up, base_dir=base_dir)
            if not self._path:
                self._path = self._tmp_dir.path

    def _prepare_db_locations(self, locations, db_day_boundary_tz_set=None, kind=None, default_location=None):
        if not locations:
            locations = [{}]
        result = []
        # set default properties if they are not specified
        for location in locations:
            location.setdefault("access_method", constants.access_method.FILE)
            location.setdefault("start_time", constants.DEFAULT_START_DATE - timedelta(days=2))
            location.setdefault("end_time", constants.DEFAULT_END_DATE + timedelta(days=1))
            if not db_day_boundary_tz_set and db_day_boundary_tz_set is not None:
                # If the day_boundary_tz is not set database-wide, then we want it to have
                # a default value for each location
                day_boundary_tz = utils.default_day_boundary_tz(self._db_suffix)
                if day_boundary_tz:
                    location.setdefault("day_boundary_tz", day_boundary_tz)

            if 'location' not in location:
                methods = {constants.access_method.SOCKET, constants.access_method.MEMORY}
                if location['access_method'] in methods:
                    raise ValueError("Parameter 'location' must be specified when parameter"
                                     f" 'access_method' is set to {methods}")
                if not default_location:
                    self._create_db()
                    location['location'] = self._path
                else:
                    location['location'] = default_location

            if kind == 'accelerator':
                location.setdefault("archive_duration", "continuous")
                location.setdefault("growable_archive", "true")
            # TODO: think what to do if there will be several locations

            result.append(location)

        return list(map(self._format_params, result))

    def _prepare_db_raw_data(self, raw_data, db_properties):
        if not raw_data:
            return []

        raw_ids = set()
        auto_discover_mounts = db_properties.get('auto_discover_mounts', '').lower() == 'yes'
        default_location = None

        for raw_db in raw_data:
            raw_db.setdefault('id', 'PRIMARY')
            if raw_db['id'] in raw_ids:
                raise ValueError("Parameter 'id' must be set and must be unique for raw databases")
            raw_ids.add(raw_db['id'])

            if 'prefix' not in raw_db:
                raise ValueError("Parameter 'prefix' must be specified for raw database")

            if self._path is not None and default_location is None:
                default_location = utils.TmpDir(rel_path='raw', base_dir=self._path, clean_up=self._clean_up).path
            raw_db['locations'] = self._prepare_db_locations(raw_db.get('locations'),
                                                             db_day_boundary_tz_set=None,
                                                             default_location=default_location)
            if auto_discover_mounts and len(raw_db['locations']) > 1:
                raise ValueError("Only one location must be specified for raw database"
                                 " when parameter 'auto_discover_mounts' is specified for database")
            for location in raw_db['locations']:
                if 'mount' not in location and not auto_discover_mounts:
                    raise ValueError("Parameter 'mount' must be specified for raw database location")
                if 'mount' in location and auto_discover_mounts:
                    raise ValueError("Parameter 'mount' must not be specified for raw database location"
                                     " when parameter 'auto_discover_mounts' is specified for database")
        return raw_data

    def _prepare_db_feed(self, feed):
        if not feed:
            return {}
        if 'type' not in feed:
            raise ValueError("Parameter 'type' must be specified for database feed")
        if feed['type'] == 'rawdb':
            feed.setdefault('format', 'native')
            formats = ('native', 'rt', 'ascii', 'xml')
            if feed['format'] not in formats:
                raise ValueError(f"Parameter 'format' must be one of {formats}")
            feed.setdefault('raw_source', 'PRIMARY')
        return self._format_params(feed)


class RefDB(DB):
    """ Creates reference database object.

    Parameters
    ----------
    name : str
        Database name
    clean_up : bool, optional
        Flag that controls temporary database cleanup
    db_properties : :obj:`dict`, optional
        Properties of database to add to locator
    db_location : :obj:`dict`, optional
        Location of database to add to locator. Reference database must have a single location,
        pointing to a continuous archive database.
    write : bool, optional
        Flag that controls access to write to database
    destroy_access : bool, optional
        Flag that controls access to destroy to database

    Examples
    --------

    >>> properties = {'symbology': 'TICKER'}
    >>> location = {'archive_duration': 'continuous'}
    >>> ref_db = otp.RefDB('REF_DATA_MYDB', db_properties=properties, db_location=location)
    >>> session.use(ref_db)
    >>>
    >>> data = 'A||20100102000000|20100103000000|B||20100103000000|20100104000000|'
    >>> out, err = ref_db.put([otp.RefDB.SymbolNameHistory(data, 'TICKER')])
    >>> b'Total ticks 8' in err and b'Total symbols 6' in err
    True
    >>>
    >>> properties = {'ref_data_db': ref_db.name, 'symbology': 'TICKER'}
    >>> db = otp.DB('MYDB', db_properties=properties)
    >>> session.use(db)
    >>>
    >>> data = otp.Ticks(X=['hello'], start=otp.dt(2010, 1, 2), end=otp.dt(2010, 1, 3))
    >>> data = otp.run(data.write(db.name, 'A', 'MSG', date=otp.dt(2010, 1, 2)))
    >>> data = otp.Ticks(X=['world!'], start=otp.dt(2010, 1, 3), end=otp.dt(2010, 1, 4))
    >>> data = otp.run(data.write(db.name, 'B', 'MSG', date=otp.dt(2010, 1, 3)))
    >>>
    >>> data = otp.DataSource(db.name, tick_type='MSG')
    >>> s_dt, e_dt, symbol_date = otp.dt(2010, 1, 1), otp.dt(2010, 1, 4), otp.dt(2010, 1, 2)
    >>> otp.run(data, symbols='A', start=s_dt, end=e_dt, symbol_date=symbol_date)
            Time       X
    0 2010-01-02   hello
    1 2010-01-03  world!
    """

    def __init__(
        self,
        name=None,
        kind='archive',
        db_properties=None,
        db_location=None,
        write=True,
        clean_up=utils.default,
        destroy_access=False,
    ):
        # ref db must have a single location, pointing to a continuous archive database
        # (its location in the locator file must have archive_duration=continuous set)
        if db_location is None:
            db_location = {}
        db_location.setdefault('archive_duration', 'continuous')
        super().__init__(
            name=name,
            kind=kind,
            db_properties=db_properties,
            db_locations=[db_location],
            write=write,
            clean_up=clean_up,
            destroy_access=destroy_access,
        )

    class Section():
        """ Specification of a reference database section. Section content can be specified as a string or source.
        The format of string and output columns of source must correspond with the section documentation.

        Parameters
        ----------
        name : str
            Section name
        data : str or :class:`otp.Source`
            Content of the section
        attrs : :obj:`dict`, optional
            Attributes of the section

        Examples
        --------

        Data provided as a string:

        >>> data = 'SYM1|20100101093000|20100101110000' + os.linesep
        >>> data += 'SYM2|20100101110000|20100103140000'
        >>> section = otp.RefDB.Section('SECTION_NAME', data, {'ATTR1': 'VAL1', 'ATTR2': 'VAL2'})
        >>> print(section)
        <SECTION_NAME ATTR1="VAL1" ATTR2="VAL2">
        SYM1|20100101093000|20100101110000
        SYM2|20100101110000|20100103140000
        </SECTION_NAME>

        Data provided as a :class:`otp.Source`:

        >>> data = dict()
        >>> data['SYMBOL_NAME'] = ['SYM1', 'SYM2']
        >>> data['START_DATETIME'] = [otp.dt(2010, 1, 1, 9, 30, tz='EST5EDT'), otp.dt(2010, 1, 1, 11, tz='EST5EDT')]
        >>> data['END_DATETIME'] = [otp.dt(2010, 1, 1, 11, tz='EST5EDT'), otp.dt(2010, 1, 3, 14, tz='EST5EDT')]
        >>> ticks = otp.Ticks(**data, offset=[0] * 2, db='LOCAL')
        >>> section = otp.RefDB.Section('SECTION_NAME', ticks, {'ATTR1': 'VAL1', 'ATTR2': 'VAL2'})
        >>> print(section) # doctest:+ELLIPSIS
        <SECTION_NAME ATTR1="VAL1" ATTR2="VAL2" OTQ_QUERY=...>
        </SECTION_NAME>

        where OTQ_QUERY is path to :class:`otp.Source`, dumped to disk as temporary .otq file.
        """
        # Read ref db guide for details on input format of sections
        # http://solutions.pages.soltest.onetick.com/iac/onetick-server/ReferenceDatabaseGuide.html
        def __init__(self, name: str, data: Union[str, 'otp.Source'], attrs: Optional[dict] = None):
            self._name = name
            self._data = data
            self._attrs = ' '.join([f'{name}="{value}"' for name, value in attrs.items()]) if attrs else ''

        def __str__(self):
            if isinstance(self._data, str):
                return f'<{self._name} {self._attrs}>{os.linesep}{self._data}{os.linesep}</{self._name}>'
            otq = self._data.to_otq()
            return f'<{self._name} {self._attrs} OTQ_QUERY={otq}>{os.linesep}</{self._name}>'

    class SymbolNameHistory(Section):
        """ Describes symbol changes for the same security. The continuity can be expressed in terms of any symbol type
        and can be specified on the security level or the security+exchange level (more explicit).

        Examples
        --------

        >>> data = 'CORE_A||20100101093000|20100101110000|CORE_B||20100101110000|20100103140000|'
        >>> section = otp.RefDB.SymbolNameHistory(data, symbology='CORE')
        >>> print(section)
        <SYMBOL_NAME_HISTORY SYMBOLOGY="CORE">
        CORE_A||20100101093000|20100101110000|CORE_B||20100101110000|20100103140000|
        </SYMBOL_NAME_HISTORY>

        Equivalent :class:`otp.Source`:

        >>> data = dict()
        >>> data['SYMBOL_NAME'] = ['CORE_A'] * 2
        >>> data['SYMBOL_NAME_IN_HISTORY'] = ['CORE_A', 'CORE_B']
        >>> data['SYMBOL_START_DATETIME'] = [otp.dt(2010, 1, 2, tz='EST5EDT')] * 2
        >>> data['SYMBOL_END_DATETIME'] = [otp.dt(2010, 1, 5, tz='EST5EDT')] * 2
        >>> data['START_DATETIME'] = [otp.dt(2010, 1, 2, tz='EST5EDT'), otp.dt(2010, 1, 3, tz='EST5EDT')]
        >>> data['END_DATETIME'] = [otp.dt(2010, 1, 3, tz='EST5EDT'), otp.dt(2010, 1, 4, tz='EST5EDT')]
        >>> ticks = otp.Ticks(**data, offset=[0] * 2, db='LOCAL')
        >>> section = otp.RefDB.SymbolNameHistory(ticks, symbology='CORE')
        >>> print(section) # doctest:+ELLIPSIS
        <SYMBOL_NAME_HISTORY SYMBOLOGY="CORE" OTQ_QUERY=...>
        </SYMBOL_NAME_HISTORY>
        """
        def __init__(self, data: Union[str, 'otp.Source'], symbology: str):
            super().__init__('SYMBOL_NAME_HISTORY', data, {'SYMBOLOGY': symbology})

    class SymbologyMapping(Section):
        """ Describes a history of mapping of symbols of one symbology to the symbols of another symbology.

        Examples
        --------

        >>> data = 'A||20100101093000|20100101110000|CORE_A|' + os.linesep
        >>> data += 'B||20100101110000|20100103140000|CORE_B|'
        >>> section = otp.RefDB.SymbologyMapping(data, source_symbology='TICKER', dest_symbology='CORE')
        >>> print(section)
        <SYMBOLOGY_MAPPING SOURCE_SYMBOLOGY="TICKER" DEST_SYMBOLOGY="CORE">
        A||20100101093000|20100101110000|CORE_A|
        B||20100101110000|20100103140000|CORE_B|
        </SYMBOLOGY_MAPPING>

        Equivalent :class:`otp.Source`:

        >>> data = dict()
        >>> data['SYMBOL_NAME'] = ['A', 'B']
        >>> data['MAPPED_SYMBOL_NAME'] = ['CORE_A', 'CORE_B']
        >>> data['START_DATETIME'] = [otp.dt(2010, 1, 2, tz='EST5EDT'), otp.dt(2010, 1, 3, tz='EST5EDT')]
        >>> data['END_DATETIME'] = [otp.dt(2010, 1, 3, tz='EST5EDT'), otp.dt(2010, 1, 4, tz='EST5EDT')]
        >>> ticks = otp.Ticks(**data, offset=[0] * 2, db='LOCAL')
        >>> section = otp.RefDB.SymbologyMapping(ticks, source_symbology='TICKER', dest_symbology='CORE')
        >>> print(section) # doctest:+ELLIPSIS
        <SYMBOLOGY_MAPPING SOURCE_SYMBOLOGY="TICKER" DEST_SYMBOLOGY="CORE" OTQ_QUERY=...>
        </SYMBOLOGY_MAPPING>
        """
        def __init__(self, data: Union[str, 'otp.Source'], source_symbology: str, dest_symbology: str):
            super().__init__('SYMBOLOGY_MAPPING', data, {'SOURCE_SYMBOLOGY': source_symbology,
                                                         'DEST_SYMBOLOGY': dest_symbology})

    class CorpActions(Section):
        """ Describes corporate actions. Used by OneTick to compute prices adjusted for various types of
        corporate actions. Supports both built-in and custom (user-defined) types of corporate actions.

        Examples
        --------

        >>> data = 'CORE_C||20100103180000|0.25|0.0|SPLIT'
        >>> section = otp.RefDB.CorpActions(data, symbology='CORE')
        >>> print(section)
        <CORP_ACTIONS SYMBOLOGY="CORE">
        CORE_C||20100103180000|0.25|0.0|SPLIT
        </CORP_ACTIONS>

        Equivalent :class:`otp.Source`:

        >>> data = dict()
        >>> data['SYMBOL_NAME'] = ['CORE_C']
        >>> data['EFFECTIVE_DATETIME'] = [otp.dt(2010, 1, 3, 18, tz='EST5EDT')]
        >>> data['MULTIPLICATIVE_ADJUSTMENT'] = [0.25]
        >>> data['ADDITIVE_ADJUSTMENT'] = [0.0]
        >>> data['ADJUSTMENT_TYPE_NAME'] = ['SPLIT']
        >>> ticks = otp.Ticks(**data, offset=[0], db='LOCAL')
        >>> section = otp.RefDB.CorpActions(ticks, symbology='CORE')
        >>> print(section) # doctest:+ELLIPSIS
        <CORP_ACTIONS SYMBOLOGY="CORE" OTQ_QUERY=...>
        </CORP_ACTIONS>
        """
        def __init__(self, data: Union[str, 'otp.Source'], symbology: str):
            super().__init__('CORP_ACTIONS', data, {'SYMBOLOGY': symbology})

    class ContinuousContracts(Section):
        """ Describes continuous contracts. Continuity is expressed in terms of stitched history
        of real contracts and rollover adjustments in between them and can be specified
        on the continuous contract level or continuous contract+exchange level (more explicit).

        Examples
        --------

        >>> data = 'CC||CORE_A||20100101093000|20100101110000|0.5|0|CORE_B||20100101110000|20100103140000'
        >>> section = otp.RefDB.ContinuousContracts(data, symbology='CORE')
        >>> print(section)
        <CONTINUOUS_CONTRACTS SYMBOLOGY="CORE">
        CC||CORE_A||20100101093000|20100101110000|0.5|0|CORE_B||20100101110000|20100103140000
        </CONTINUOUS_CONTRACTS>

        Equivalent :class:`otp.Source`:

        >>> data = dict()
        >>> data['CONTINUOUS_CONTRACT_NAME'] = ['CC'] * 2
        >>> data['SYMBOL_NAME'] = ['CORE_A', 'CORE_B']
        >>> data['START_DATETIME'] = [otp.dt(2010, 1, 2, tz='EST5EDT'), otp.dt(2010, 1, 3, tz='EST5EDT')]
        >>> data['END_DATETIME'] = [otp.dt(2010, 1, 3, tz='EST5EDT'), otp.dt(2010, 1, 4, tz='EST5EDT')]
        >>> data['MULTIPLICATIVE_ADJUSTMENT'] = [0.5, None]
        >>> data['ADDITIVE_ADJUSTMENT'] = [3, None]
        >>> ticks = otp.Ticks(**data, offset=[0] * 2, db='LOCAL')
        >>> section = otp.RefDB.ContinuousContracts(ticks, symbology='CORE')
        >>> print(section) # doctest:+ELLIPSIS
        <CONTINUOUS_CONTRACTS SYMBOLOGY="CORE" OTQ_QUERY=...>
        </CONTINUOUS_CONTRACTS>
        """
        def __init__(self, data: Union[str, 'otp.Source'], symbology: str):
            super().__init__('CONTINUOUS_CONTRACTS', data, {'SYMBOLOGY': symbology})

    class SymbolCurrency(Section):
        """ Specifies symbols' currencies in 3-letter ISO codes for currencies. These are used for currency conversion
        (e.g., when calculating portfolio price for a list of securities with different currencies).

        Examples
        --------

        >>> data = 'CORE_A||20100101093000|20100101110000|USD|1.0' + os.linesep
        >>> data += 'CORE_B||20100101110000|20100103140000|RUB|1.8'
        >>> section = otp.RefDB.SymbolCurrency(data, symbology='CORE')
        >>> print(section)
        <SYMBOL_CURRENCY SYMBOLOGY="CORE">
        CORE_A||20100101093000|20100101110000|USD|1.0
        CORE_B||20100101110000|20100103140000|RUB|1.8
        </SYMBOL_CURRENCY>

        Equivalent :class:`otp.Source`:

        >>> data = dict()
        >>> data['SYMBOL_NAME'] = ['CORE_A', 'CORE_B',]
        >>> data['CURRENCY'] = ['USD', 'RUB']
        >>> data['MULTIPLIER'] = [1., 1.8]
        >>> data['START_DATETIME'] = [otp.dt(2010, 1, 1, 9, 30, tz='EST5EDT'), otp.dt(2010, 1, 1, 11, tz='EST5EDT')]
        >>> data['END_DATETIME'] = [otp.dt(2010, 1, 1, 11, tz='EST5EDT'), otp.dt(2010, 1, 3, 14, tz='EST5EDT')]
        >>> ticks = otp.Ticks(**data, offset=[0] * 2, db='LOCAL')
        >>> section = otp.RefDB.SymbolCurrency(ticks, symbology='CORE')
        >>> print(section) # doctest:+ELLIPSIS
        <SYMBOL_CURRENCY SYMBOLOGY="CORE" OTQ_QUERY=...>
        </SYMBOL_CURRENCY>
        """
        def __init__(self, data: Union[str, 'otp.Source'], symbology: str):
            super().__init__('SYMBOL_CURRENCY', data, {'SYMBOLOGY': symbology})

    class Calendar(Section):
        """ Specifies a named calendar. Needed to analyze tick data during specific market time intervals (i.e., during
        normal trading hours). Can either be used directly in queries as described below, or referred to
        from the SYMBOL_CALENDAR and EXCH_CALENDAR sections.

        Examples
        --------

        >>> data = 'CAL1|20100101093000|20100101110000|Regular|R|0.0.12345|093000|160000|GMT|1|DESCRIPTION1'
        >>> data += os.linesep
        >>> data += 'CAL2|20100101110000|20100103140000|Holiday|F|0.0.12345|094000|170000|GMT|0|DESCRIPTION2'
        >>> section = otp.RefDB.Calendar(data)
        >>> print(section)
        <CALENDAR >
        CAL1|20100101093000|20100101110000|Regular|R|0.0.12345|093000|160000|GMT|1|DESCRIPTION1
        CAL2|20100101110000|20100103140000|Holiday|F|0.0.12345|094000|170000|GMT|0|DESCRIPTION2
        </CALENDAR>

        Equivalent :class:`otp.Source`:

        >>> data = dict()
        >>> data['CALENDAR_NAME'] = ['CAL1', 'CAL2']
        >>> data['START_DATETIME'] = [otp.dt(2010, 1, 1, 9, 30, tz='EST5EDT'), otp.dt(2010, 1, 1, 11, tz='EST5EDT')]
        >>> data['END_DATETIME'] = [otp.dt(2010, 1, 1, 11, tz='EST5EDT'), otp.dt(2010, 1, 3, 14, tz='EST5EDT')]
        >>> data['SESSION_NAME'] = ['Regular', 'Holiday']
        >>> data['SESSION_FLAGS'] = ['R', 'H']
        >>> data['DAY_PATTERN'] = ['0.0.12345', '0.0.12345']
        >>> data['START_HHMMSS'] = ['093000', '094000']
        >>> data['END_HHMMSS'] = ['160000', '170000']
        >>> data['TIMEZONE'] = ['GMT', 'GMT']
        >>> data['PRIORITY'] = [1, 0]
        >>> data['DESCRIPTION'] = ['DESCRIPTION1', 'DESCRIPTION2']
        >>> ticks = otp.Ticks(**data, offset=[0] * 2, db='LOCAL')
        >>> section = otp.RefDB.Calendar(ticks)
        >>> print(section) # doctest:+ELLIPSIS
        <CALENDAR  OTQ_QUERY=...>
        </CALENDAR>
        """
        def __init__(self, data: Union[str, 'otp.Source']):
            super().__init__('CALENDAR', data)

    class SymbolCalendar(Section):
        """ Specifies a calendar for a symbol. Needed to analyze tick data during specific market time intervals
        (i.e., during normal trading hours). Can either be specified directly or refer to a named calendar by its name
        (see the CALENDAR section).

        Examples
        --------

        Symbol calendar section, referring to named calendar section:

        >>> data = 'CORE_A|20100101093000|20100101110000|CAL1' + os.linesep
        >>> data += 'CORE_B|20100101110000|20100103140000|CAL2'
        >>> section = otp.RefDB.SymbolCalendar(data, symbology='CORE')
        >>> print(section)
        <SYMBOL_CALENDAR SYMBOLOGY="CORE">
        CORE_A|20100101093000|20100101110000|CAL1
        CORE_B|20100101110000|20100103140000|CAL2
        </SYMBOL_CALENDAR>

        Equivalent :class:`otp.Source`:

        >>> data = dict()
        >>> data['SYMBOL_NAME'] = ['CORE_A', 'CORE_B']
        >>> data['START_DATETIME'] = [otp.dt(2010, 1, 1, 9, 30, tz='EST5EDT'), otp.dt(2010, 1, 1, 11, tz='EST5EDT')]
        >>> data['END_DATETIME'] = [otp.dt(2010, 1, 1, 11, tz='EST5EDT'), otp.dt(2010, 1, 3, 14, tz='EST5EDT')]
        >>> data['CALENDAR_NAME'] = ['CAL1', 'CAL2']
        >>> ticks = otp.Ticks(**data, offset=[0] * 2, db='LOCAL')
        >>> section = otp.RefDB.SymbolCalendar(ticks, symbology='CORE')
        >>> print(section) # doctest:+ELLIPSIS
        <SYMBOL_CALENDAR SYMBOLOGY="CORE" OTQ_QUERY=...>
        </SYMBOL_CALENDAR>

        Symbol calendar section without using named calendar section:

        >>> data = 'CORE_A|20100101093000|20100101110000|Regular|R|0.0.12345|093000|160000|EST5EDT|1|' + os.linesep
        >>> data += 'CORE_B|20100101110000|20100103140000|Regular|F|0.0.12345|093000|160000|EST5EDT|1|'
        >>> section = otp.RefDB.SymbolCalendar(data, symbology='CORE')
        >>> print(section)
        <SYMBOL_CALENDAR SYMBOLOGY="CORE">
        CORE_A|20100101093000|20100101110000|Regular|R|0.0.12345|093000|160000|EST5EDT|1|
        CORE_B|20100101110000|20100103140000|Regular|F|0.0.12345|093000|160000|EST5EDT|1|
        </SYMBOL_CALENDAR>

        Equivalent :class:`otp.Source`:

        >>> data = dict()
        >>> data['SYMBOL_NAME'] = ['CORE_A', 'CORE_B']
        >>> data['START_DATETIME'] = [otp.dt(2010, 1, 1, 9, 30, tz='EST5EDT'), otp.dt(2010, 1, 1, 11, tz='EST5EDT')]
        >>> data['END_DATETIME'] = [otp.dt(2010, 1, 1, 11, tz='EST5EDT'), otp.dt(2010, 1, 3, 14, tz='EST5EDT')]
        >>> data['SESSION_NAME'] = ['Regular', 'Regular']
        >>> data['SESSION_FLAGS'] = ['R', 'F']
        >>> data['DAY_PATTERN'] = ['0.0.12345', '0.0.12345']
        >>> data['START_HHMMSS'] = ['093000', '160000']
        >>> data['END_HHMMSS'] = ['CAL1', 'CAL2']
        >>> data['TIMEZONE'] = ['EST5EDT', 'EST5EDT']
        >>> data['PRIORITY'] = [1, 1]
        >>> data['DESCRIPTION'] = ['', '']
        >>> ticks = otp.Ticks(**data, offset=[0] * 2, db='LOCAL')
        >>> section = otp.RefDB.SymbolCalendar(ticks, symbology='CORE')
        >>> print(section) # doctest:+ELLIPSIS
        <SYMBOL_CALENDAR SYMBOLOGY="CORE" OTQ_QUERY=...>
        </SYMBOL_CALENDAR>
        """
        def __init__(self, data: Union[str, 'otp.Source'], symbology: str):
            super().__init__('SYMBOL_CALENDAR', data, {'SYMBOLOGY': symbology})

    class SectionStr(Section):
        """ Specification of a reference database section that can be specified only as a string.
        Section content still can be provided as a :class:`otp.Source`, but the :class:`otp.Source` is executed and
        result data is used as string in section. It's up to user to provide :class:`otp.Source` with correct number
        and order of columns.

        Examples
        --------

        Data provided as a string returns the same result as :class:`otp.RefDB.Section`.

        Data provided as a :class:`otp.Source`:

        >>> data = dict()
        >>> data['SYMBOL_NAME'] = ['SYM1', 'SYM2']
        >>> data['START_DATETIME'] = [otp.dt(2010, 1, 1, 9, 30, tz='EST5EDT'), otp.dt(2010, 1, 1, 11, tz='EST5EDT')]
        >>> data['END_DATETIME'] = [otp.dt(2010, 1, 1, 11, tz='EST5EDT'), otp.dt(2010, 1, 3, 14, tz='EST5EDT')]
        >>> ticks = otp.Ticks(**data, offset=[0] * 2, db='LOCAL')
        >>> ticks = ticks.table(SYMBOL_NAME=otp.string[128], START_DATETIME=otp.msectime, END_DATETIME=otp.msectime)
        >>> section = otp.RefDB.SectionStr('SECTION_NAME', ticks, {'ATTR1': 'VAL1', 'ATTR2': 'VAL2'})
        >>> print(section) # doctest:+ELLIPSIS
        <SECTION_NAME ATTR1="VAL1" ATTR2="VAL2">
        SYM1|20100101093000|20100101110000
        SYM2|20100101110000|20100103140000
        </SECTION_NAME>

        where OTQ_QUERY is path to :class:`otp.Source`, dumped to disk as temporary .otq file.
        """
        def __init__(self, name: str, data: Union[str, 'otp.Source'], attrs: Optional[dict] = None):
            data_str = data if isinstance(data, str) else self._source_to_str(data)
            super().__init__(name, data_str, attrs)

        def _source_to_str(self, data: 'otp.Source'):
            df = otp.run(data)
            df.drop(columns=['Time'], inplace=True)
            df = df.to_csv(sep='|', header=False, index=False, date_format='%Y%m%d%H%M%S')
            return df

    class PrimaryExchange(SectionStr):
        """ Specifies symbols' primary exchanges. Used to extract and analyze tick data for a security on
        its primary exchange, without having to explicitly specify the name of the primary exchange.

        Examples
        --------

        >>> data = 'A||19991118000000|99999999000000|N|'
        >>> data += os.linesep
        >>> data += 'AA||19991118000000|99999999000000|N|AA.N'
        >>> section = otp.RefDB.PrimaryExchange(data, symbology='TICKER')
        >>> print(section)
        <PRIMARY_EXCHANGE SYMBOLOGY="TICKER">
        A||19991118000000|99999999000000|N|
        AA||19991118000000|99999999000000|N|AA.N
        </PRIMARY_EXCHANGE>

        Equivalent query should return the same data values in the same order. Column names does not matter.
        """
        def __init__(self, data: Union[str, 'otp.Source'], symbology: str):
            super().__init__('PRIMARY_EXCHANGE', data, {'SYMBOLOGY': symbology})

    class ExchCalendar(SectionStr):
        """ Specifies symbols' primary exchanges. Used to extract and analyze tick data for a security on
        its primary exchange, without having to explicitly specify the name of the primary exchange.

        Examples
        --------

        >>> data = 'NYSE||19600101000000|20501231235959|Regular|R|0.0.12345|093000|160000|EST5EDT|'
        >>> data += os.linesep
        >>> data += 'NYSE||19600101000000|20501231235959|Half-day|RL|12/31|093000|130000|EST5EDT|'
        >>> data += os.linesep
        >>> data += 'NYSE||19600101000000|20501231235959|Holiday|H|01/01|000000|240000|EST5EDT|'
        >>> section = otp.RefDB.ExchCalendar(data, symbology='MIC')
        >>> print(section)
        <EXCH_CALENDAR SYMBOLOGY="MIC">
        NYSE||19600101000000|20501231235959|Regular|R|0.0.12345|093000|160000|EST5EDT|
        NYSE||19600101000000|20501231235959|Half-day|RL|12/31|093000|130000|EST5EDT|
        NYSE||19600101000000|20501231235959|Holiday|H|01/01|000000|240000|EST5EDT|
        </EXCH_CALENDAR>

        If a CALENDAR section is used:

        >>> data = 'LSE||19600101000000|20501231235959|WNY'
        >>> section = otp.RefDB.ExchCalendar(data, symbology='MIC')
        >>> print(section)
        <EXCH_CALENDAR SYMBOLOGY="MIC">
        LSE||19600101000000|20501231235959|WNY
        </EXCH_CALENDAR>

        Equivalent query should return the same data values in the same order. Column names does not matter.
        """
        def __init__(self, data: Union[str, 'otp.Source'], symbology: str):
            super().__init__('EXCH_CALENDAR', data, {'SYMBOLOGY': symbology})

    class SymbolExchange(SectionStr):
        """ Specifies the exchange where a security is traded. Needs to be provided for the symbologies where
        the symbol name is unique across all exchanges.

        Examples
        --------

        >>> data = 'IBM.N|19980825000000|20501231235959|NYSE||'
        >>> section = otp.RefDB.SymbolExchange(data, symbology='RIC', exchange_symbology='MIC')
        >>> print(section)
        <SYMBOL_EXCHANGE SYMBOLOGY="RIC" EXCHANGE_SYMBOLOGY="MIC">
        IBM.N|19980825000000|20501231235959|NYSE||
        </SYMBOL_EXCHANGE>

        Equivalent query should return the same data values in the same order. Column names does not matter.
        """
        def __init__(self, data: Union[str, 'otp.Source'], symbology: str, exchange_symbology: str):
            super().__init__('SYMBOL_EXCHANGE', data, {'SYMBOLOGY': symbology,
                                                       'EXCHANGE_SYMBOLOGY': exchange_symbology})

    def put(
        self,
        src: Union[str, List[Section]],
        tickdb_symbology: Optional[List[str]] = None,
        delta_mode: bool = False,
        full_integrity_check: bool = False,
        load_by_sections: bool = True,
    ):
        """
        Loads data in database with reference_data_loader.exe.
        If db properties contain SUPPORT_DELTAS=YES, delta_mode set to True, and proper delta file is used
        then data is loaded in incremental mode (in other words, replace or modification mode).
        If the above conditions are not met, reference database content is entirely rewritten with the new data.

        Parameters
        ----------
        src : str, list of str or otp.RefDB.Section
            Path to data file, or list of data per section in specified format
        tickdb_symbology : list of str, optional
            All symbologies for which the reference data needs to be generated
        delta_mode : bool, default is False
            If set to True loader will perform incremental load. Cannot be used if ``tickdb_symbology`` is specified
        full_integrity_check : bool, default is False
            If set to True loader checks all mappings to symbologies with symbol name history section
            and gives warning if mapped securities do not have symbol name history
        load_by_sections : bool, default is True
            If set to True loader will perform input data file splitting by data types and symbologies
            to load each part separately instead loading the entire file at once
        """
        # More info:
        # http://solutions.pages.soltest.onetick.com/iac/onetick-server/reference_data_loader.html - loader doc
        # https://onemarketdata.atlassian.net/browse/KB-286 - details on delta mode
        return self._session_handler(
            self._put,
            src=src,
            delta_mode=delta_mode,
            full_integrity_check=full_integrity_check,
            load_by_sections=load_by_sections,
            tickdb_symbology=tickdb_symbology,
        )

    def _prepare_data_file(self, src):
        if isinstance(src, str):
            return src

        data = f'<VERSION_INFO VERSION="1">{os.linesep}</VERSION_INFO>'
        for section in src:
            data += f'{os.linesep}{os.linesep}{section}'

        data_file = utils.TmpFile(suffix='.txt')
        with open(data_file.path, 'w') as f:
            f.writelines(data)
        return data_file.path

    def _prepare_loader_args(self, data_file, tickdb_symbology, delta_mode, full_integrity_check, load_by_sections):
        loader_args = ['-dbname', self.name]
        loader_args += ['-data_file', data_file]
        if tickdb_symbology:
            for symbology in tickdb_symbology:
                loader_args += ['-tickdb_symbology', symbology]
        loader_args += ['-delta_mode', 'yes' if delta_mode else 'no']
        loader_args += ['-full_integrity_check', 'yes' if full_integrity_check else 'no']
        loader_args += ['-load_by_sections', 'yes' if load_by_sections else 'no']
        return loader_args

    def _put(self, src, tickdb_symbology, delta_mode, full_integrity_check, load_by_sections):
        if otp.__webapi__:
            raise NotImplementedError("Reference database loader is not supported in WebAPI mode.")
        data_file = self._prepare_data_file(src)
        loader_args = self._prepare_loader_args(
            data_file, tickdb_symbology, delta_mode, full_integrity_check, load_by_sections
        )
        loader_path = os.path.join(utils.omd_dist_path(), 'one_tick', 'bin', 'reference_data_loader.exe')
        p = subprocess.run(
            [loader_path] + loader_args,
            env={
                'ONE_TICK_CONFIG': session.Session._instance.config.path,
                'TZ': os.environ.get('TZ', otp.config['tz']),
            },
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=False,
        )
        return p.stdout, p.stderr

    def add(self, *args, **kwargs):
        # this method is not implemented because
        # reference database loader can only rewrite the data, not add new entries
        raise NotImplementedError("Method is not supported for reference databases. Use put instead.")



import os

from functools import partial

import onetick.py as otp
from onetick.py.otq import otq

from onetick.py.core.source import Source

from .. import utils

from .common import update_node_tick_type


class SplitQueryOutputBySymbol(Source):
    def __init__(self,
                 query=None,
                 symbol_field=None,
                 single_invocation=False,
                 db=utils.adaptive_to_default,
                 tick_type=utils.adaptive,
                 start=utils.adaptive,
                 end=utils.adaptive,
                 symbols=utils.adaptive,
                 schema=None,
                 **kwargs):
        """
        A data source used to dispatch output ticks,
        resulting after execution of the specified query,
        according to the values of the specified field in those ticks.

        Each replica of this EP, corresponding to a particular bound or unbound symbol, thus,
        propagates resulting ticks of the specified query,
        with values of the specified field in those ticks equal to that symbol.

        Note, that database name part of symbols is not taken into account.

        Parameters
        ----------
        query:
            Specify query to execute.
        symbol_field:
            Specifies the field in the resulting ticks of the underlying query,
            according to values of which those ticks are dispatched.
        single_invocation:
            By default, the underlying query is executed once
            per symbol batch and per execution thread of the containing query.
            If this parameter is set to True, the underlying query is executed once
            regardless of the batch size and the number of CPU cores utilized.

            The former option should be the preferred (hence the default) one, as it reduces memory overhead,
            while the latter one might be chosen to speed-up the overall execution.

            Note, that if the underlying query is a CEP query, than this option has no effect,
            as there is a single batch and a single thread anyway.
        db:
        tick_type:
            Tick type to set on the OneTick's graph node.
            Can be used to specify database name with tick type or tick type only.

            By default setting these parameters is not required, database is usually set
            with parameter ``symbols`` or in :py:func:`otp.run <onetick.py.run>`.
        start:
            Custom start time of the source.
            If set, will override the value specified in :py:func:`otp.run <onetick.py.run>`.
        end:
            Custom end time of the source.
            If set, will override the value specified in :py:func:`otp.run <onetick.py.run>`.
        symbols:
            Symbol(s) from which data should be taken.
            If set, will override the value specified in :py:func:`otp.run <onetick.py.run>`.

        Examples
        --------

        Get only the ticks that have needed symbols specified in field ``TICKER``:

        >>> data = otp.Ticks(X=[1, 2, 3, 4], TICKER=['A', 'B', 'A', 'C'])
        >>> data = otp.SplitQueryOutputBySymbol(data, data['TICKER'])
        >>> res = otp.run(data, symbols=['A', 'B'])
        >>> res['A']
                             Time  X TICKER
        0 2003-12-01 00:00:00.000  1      A
        1 2003-12-01 00:00:00.002  3      A
        >>> res['B']
                             Time  X TICKER
        0 2003-12-01 00:00:00.001  2      B
        """

        if self._try_default_constructor(schema=schema, **kwargs):
            return

        if isinstance(query, Source):
            query = query.copy()
            otq_query = query.to_otq()
            q_start, q_end, _ = query._set_date_range_and_symbols()
            if start is utils.adaptive and end is utils.adaptive:
                start, end = q_start, q_end
        else:
            # TODO: support already existing queries
            raise ValueError('Non supported type of the `query` is specified')

        super().__init__(
            _symbols=symbols,
            _start=start,
            _end=end,
            _base_ep_func=partial(self.build, db, tick_type, symbol_field, otq_query, single_invocation),
            schema=schema,
            **kwargs,
        )

    def build(self, db, tick_type, symbol_field_name, otq_query, single_invocation):
        src = Source(otq.SplitQueryOutputBySymbol(otq_query=otq_query,
                                                  symbol_field_name=str(symbol_field_name),
                                                  ensure_single_invocation=single_invocation))

        update_node_tick_type(src, tick_type, db)

        return src


def by_symbol(src: Source,
              symbol_field,
              single_invocation=False,
              db=utils.adaptive_to_default,
              tick_type=utils.adaptive,
              start=utils.adaptive,
              end=utils.adaptive,
              ) -> Source:
    """
    Create a separate data series for each unique value of ``symbol_field`` in the output of ``src``.
    ``src`` must specify enough parameters to be run (e.g., symbols, query range). A typical use case is to split a
    single data series (e.g., from a CSV file) into separate data series by symbol. This method is a source.

    Parameters
    ----------
    src: Source
        a query which output is to be split by ``symbol_field``
    symbol_field: str
        the name of the field carrying symbol name in the ``src`` query
    single_invocation: bool, optional
        ``True`` means that the ``src`` query is run once and the result stored in memory speeding up the execution.
        ``False`` means that the ``src`` query is run for every symbol of the query saving memory
        but slowing down query execution.
        Default: ``False``
    db: str, optional
        Database for running the query. Doesn't affect the ``src`` query. The default value
        is ``otp.config['default_db']``.
    tick_type: str, optional
        Tick type for the query. Doesn't affect the ``src`` query.
    start: otp.dt, optional
        By default it is taken from the ``src`` start time
    end: otp.dt, optional
        By default it is taken from the ``src`` end time

    See also
    --------
    **SPLIT_QUERY_OUTPUT_BY_SYMBOL** OneTick event processor

    Examples
    --------
    >>> executions = otp.CSV( # doctest: +SKIP
    ...     otp.utils.file(os.path.join(cur_dir, 'data', 'example_events.csv')),
    ...     converters={"time_number": lambda c: c.apply(otp.nsectime)},
    ...     timestamp_name="time_number",
    ...     start=otp.dt(2022, 7, 1),
    ...     end=otp.dt(2022, 7, 2),
    ...     order_ticks=True
    ... )[['stock', 'px']]
    >>> csv = otp.by_symbol(executions, 'stock') # doctest: +SKIP
    >>> trd = otp.DataSource( # doctest: +SKIP
    ...     db='US_COMP',
    ...     tick_type='TRD',
    ...     start=otp.dt(2022, 7, 1),
    ...     end=otp.dt(2022, 7, 2)
    ... )[['PRICE', 'SIZE']]
    >>> data = otp.join_by_time([csv, trd]) # doctest: +SKIP
    >>> result = otp.run(data, symbols=executions.distinct(keys='stock')[['stock']], concurrency=8) # doctest: +SKIP
    >>> result['THG'] # doctest: +SKIP
                               Time stock      px   PRICE  SIZE
    0 2022-07-01 11:37:56.432947200   THG  148.02  146.48     1
    >>> result['TFX'] # doctest: +SKIP
                               Time stock      px   PRICE  SIZE
    0 2022-07-01 11:39:45.882808576   TFX  255.61  251.97     1
    >>> result['BURL'] # doctest: +SKIP
                               Time stock      px   PRICE  SIZE
    0 2022-07-01 11:42:35.125718016  BURL  137.53  135.41     2
    """
    result = SplitQueryOutputBySymbol(src,
                                      symbol_field=symbol_field,
                                      single_invocation=single_invocation,
                                      db=db,
                                      tick_type=tick_type,
                                      start=start,
                                      end=end)

    result.schema.set(**src.schema)
    return result



def get_type_that_includes(types):
    import onetick.py.types as ott

    def merge_two(type1, type2):
        type_changed = False

        b_type1, b_type2 = ott.get_base_type(type1), ott.get_base_type(type2)
        if b_type1 != b_type2:
            if {b_type1, b_type2} == {int, float}:
                dtype = float
            elif {b_type1, b_type2} == {ott.decimal, float} or {b_type1, b_type2} == {ott.decimal, int}:
                dtype = ott.decimal
            elif {b_type1, b_type2} == {ott.nsectime, ott.msectime}:
                dtype = ott.nsectime
            else:
                raise ValueError(f"Incompatible types: {type1}, {type2}")

            type_changed = True
        elif issubclass(b_type1, str):
            t1_length = ott.string.DEFAULT_LENGTH if type1 is str or type1.length is None else type1.length
            t2_length = ott.string.DEFAULT_LENGTH if type2 is str or type2.length is None else type2.length

            if t1_length is Ellipsis or t2_length is Ellipsis:
                dtype = ott.varstring
            else:
                dtype = type2 if t1_length < t2_length else type1

            if t1_length != t2_length:
                type_changed = True  # TODO: test

        else:
            dtype = type1

        return dtype, type_changed

    dtype = types[0]
    type_changed = False

    for other_dtype in types[1:]:
        dtype, new_type_change = merge_two(dtype, other_dtype)
        type_changed |= new_type_change

    return dtype, type_changed


class adaptive:
    """
    This class is mostly used as the default value for the functions' parameters
    when the value of ``None`` has some other meaning
    or when the meaning of the parameter depends on the other parameter's values,
    :ref:`otp.config <static/configuration/root:configuration>` options or the context.

    Examples
    --------

    For example, setting :py:class:`~onetick.py.DataSource` ``symbols`` parameter
    to ``otp.adaptive`` allows to set symbols when running the query later.

    >>> data = otp.DataSource('SOME_DB', tick_type='TT', symbols=otp.adaptive)
    >>> otp.run(data, symbols='S1')
                         Time  X
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.001  2
    2 2003-12-01 00:00:00.002  3

    This is the default value of ``symbols`` parameter, so omitting it also works:

    >>> data = otp.DataSource('SOME_DB', tick_type='TT')
    >>> otp.run(data, symbols='S1')
                         Time  X
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.001  2
    2 2003-12-01 00:00:00.002  3
    """


class adaptive_to_default(adaptive):
    """
    If something is not specified and can not be deduced, then use the
    default one
    """


class default:
    """
    Used when you need to specify a default without evaluating it (e.g. default timezone)
    """


class range:
    """
    Class that expresses OneTick ranges.
    For example, if you want to express a range in the .split() method,
    then you can use this range.

    It has start and stop fields that allow you to define a range.

    See also
    --------
    :py:meth:`~onetick.py.Source.split`.

    Examples
    --------
    >>> data = otp.Ticks(X=[0.33, -5.1, otp.nan, 9.4])
    >>> r1, r2, r3 = data.split(data['X'], [otp.nan, otp.range(0, 100)], default=True)
    >>> otp.run(r1)
                         Time   X
    0 2003-12-01 00:00:00.002 NaN
    >>> otp.run(r2)
                         Time     X
    0 2003-12-01 00:00:00.000  0.33
    1 2003-12-01 00:00:00.003  9.40
    >>> otp.run(r3)
                         Time    X
    0 2003-12-01 00:00:00.001 -5.1
    """

    def __init__(self, start, stop):
        self.start = start
        self.stop = stop



from typing import TYPE_CHECKING, Optional, Union

from onetick.py import types as ott
from onetick.py.core.column_operations.base import _Operation
from onetick.py.otq import otq

from .misc import inplace_operation

if TYPE_CHECKING:
    import onetick.py as otp
    from onetick.py.core.source import Source


@inplace_operation
def show_symbol_name_in_db(self: 'Source', inplace=False) -> Optional['Source']:
    """
    Adds the **SYMBOL_NAME_IN_DB** field to input ticks,
    indicating the symbol name of the tick in the database.

    Parameters
    ----------
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing.
        Otherwise method returns a new modified object.

    See also
    --------
    **SHOW_SYMBOL_NAME_IN_DB** OneTick event processor

    Returns
    -------
    :class:`Source` or ``None``

    Examples
    --------

    For example, it can be used to display
    the actual symbol name for the contract (e.g., **ESM23**)
    instead of the artificial continuous name **ES_r_tdi**.
    Notice how actual symbol name can be different for each tick,
    e.g. in this case it is different for each month.

    >>> data = otp.DataSource('TDI_FUT', tick_type='TRD')  # doctest: +SKIP
    >>> data = data[['PRICE']]  # doctest: +SKIP
    >>> data = data.first(bucket_interval=31, bucket_units='days')  # doctest: +SKIP
    >>> data['SYMBOL_NAME'] = data.Symbol.name  # doctest: +SKIP
    >>> data = data.show_symbol_name_in_db()  # doctest: +SKIP
    >>> otp.run(data,  # doctest: +SKIP
    ...         symbols='ES_r_tdi', symbol_date=otp.dt(2023, 3, 1),
    ...         start=otp.dt(2023, 3, 1), end=otp.dt(2023, 5, 1))
                         Time    PRICE SYMBOL_NAME SYMBOL_NAME_IN_DB
    0 2023-03-01 00:00:00.549  3976.75    ES_r_tdi             ESH23
    1 2023-04-02 18:00:00.039  4127.00    ES_r_tdi             ESM23
    """
    if 'SYMBOL_NAME_IN_DB' in self.schema:
        raise ValueError("Column 'SYMBOL_NAME_IN_DB' already exists.")
    self.sink(otq.ShowSymbolNameInDb())
    self.schema['SYMBOL_NAME_IN_DB'] = str
    return self


@inplace_operation
def modify_symbol_name(
    self: 'Source', symbol_name: Union[str, 'otp.Operation', 'otp.Column'], inplace=False
) -> Optional['Source']:
    """
    Modifies the name of the symbol that provides input ticks for this node.
    Uses MODIFY_SYMBOL_NAME EP.

    Parameters
    ----------
    symbol_name: str, :py:class:`~onetick.py.Column`, :py:class:`~onetick.py.Operation`
        String or expression with new `SYMBOL_NAME` value.
        New `SYMBOL_NAME` must not depend on ticks, if set via expression.
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing.
        Otherwise method returns a new modified object.

    See also
    --------
    **MODIFY_SYMBOL_NAME** OneTick event processor

    Returns
    -------
    :class:`Source` or ``None``

    Examples
    --------
    Replacing with static string:

    >>> data = otp.DataSource('SOME_DB', symbol='S1', tick_type='TT')
    >>> data = data.modify_symbol_name(symbol_name='S2')
    >>> otp.run(data)
                         Time   X
    0 2003-12-01 00:00:00.000  -3
    1 2003-12-01 00:00:00.001  -2
    2 2003-12-01 00:00:00.002  -1

    Replacing with expression:

    >>> data = otp.DataSource('SOME_DB', symbol='S2', tick_type='TT')
    >>> data = data.modify_symbol_name(symbol_name=data['_SYMBOL_NAME'].str.replace('2', '1'))
    >>> otp.run(data)
                         Time  X
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.001  2
    2 2003-12-01 00:00:00.002  3

    """
    if not isinstance(symbol_name, (str, _Operation)):
        raise ValueError("Unsupported symbol_name argument value type")

    self.sink(otq.ModifySymbolName(symbol_name=ott.value2str(symbol_name)))

    return self
